EMBED_PROVIDER_API_KEY_VARS = {
    # Embedding provider supported BYOK (Bring Your Own Key)
    "openai": "TIDB_EXP_EMBED_OPENAI_API_KEY",
    "jina_ai": "TIDB_EXP_EMBED_JINA_AI_API_KEY",
    "cohere": "TIDB_EXP_EMBED_COHERE_API_KEY",
    "huggingface": "TIDB_EXP_EMBED_HUGGINGFACE_API_KEY",
    "nvidia_nim": "TIDB_EXP_EMBED_NVIDIA_NIM_API_KEY",
    "gemini": "TIDB_EXP_EMBED_GEMINI_API_KEY",
}
